import random
import shutil

import pytest

from tests.helpers.api import ConnectorApi
from tests.helpers.log_helper import log_fixture
from tests.helpers.server_helper import ReindexerServer
from tests.test_data.constants import composite_index_definition, index_definition, item_definition


def pytest_addoption(parser):
    parser.addoption("--mode", choices=["builtin", "cproto"], default="builtin", help="Connection mode")
    parser.addoption("--rx_bin_path", action="store", help="path to reindexer_server", default="")


@pytest.fixture(scope="session", autouse=True)
def log_setup(request):
    log_fixture.info("Work with pyreindexer connector using {} mode".format(request.config.getoption("--mode")))


@pytest.fixture(scope="session", autouse=True)
def rx_server(request):
    """
    Start reindexer server for cproto mode
    """
    if request.config.getoption("--mode") == "builtin":
        yield
    else:
        rx_bin_path = request.config.getoption("--rx_bin_path")
        server = ReindexerServer(rx_bin_path=rx_bin_path, http_port=9088, rpc_port=6534, storage="/tmp/reindex_test")
        server.run()
        yield
        server.terminate()


@pytest.fixture(scope="session")
def db(request):
    """
    Create a database
    """
    mode = request.config.getoption("--mode")
    prefix = "builtin://tmp/" if mode == "builtin" else "cproto://127.0.0.1:6534/"
    db_name = "test_db"
    db = ConnectorApi(f"{prefix}{db_name}")
    yield db
    db.close()
    shutil.rmtree("tmp/", ignore_errors=True)


@pytest.fixture
def namespace(db):
    """
    Create a namespace
    """
    ns_name = 'new_ns'
    db.namespace.open(ns_name)
    yield ns_name
    db.namespace.drop(ns_name)


@pytest.fixture
def index(db, namespace):
    """
    Create an index to namespace
    """
    db.index.create(namespace, index_definition)
    yield


@pytest.fixture
def sparse_index(db, namespace):
    """
    Create sparse index to namespace
    """
    db.index.create(namespace, {"name": "val", "json_paths": ["val"], "field_type": "string", "index_type": "hash",
                                "is_sparse": True})
    yield


@pytest.fixture
def indexes(db, namespace):
    """
    Create two indexes to namespace
    """
    db.index.create(namespace, index_definition)
    db.index.create(namespace, {"name": "val", "json_paths": ["val"], "field_type": "string", "index_type": "hash"})
    yield


@pytest.fixture
def item(db, namespace):
    """
    Create an item to namespace
    """
    db.item.insert(namespace, item_definition)
    yield item_definition


@pytest.fixture
def items(db, namespace):
    """
    Create items to namespace
    """
    items = [{"id": i, "val": f"testval{i}"} for i in range(10)]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def items_shuffled(db, namespace):
    """
    Create items in random order
    """
    items = [{"id": i, "val": f"testval{i}"} for i in range(5)]
    random.shuffle(items)
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def composite_index(db, namespace):
    """
    Create indexes and composite index from them
    """
    db.index.create(namespace, index_definition)
    db.index.create(namespace, {"name": "val", "json_paths": ["val"], "field_type": "string", "index_type": "hash"})
    db.index.create(namespace, composite_index_definition)
    yield


@pytest.fixture
def rtree_index_and_items(db, namespace):
    """
    Create rtree index and items
    """
    db.index.create(namespace, {"name": "rtree", "json_paths": ["rtree"], "field_type": "point",
                                "index_type": "rtree", "rtree_type": "rstar"})
    items = [{"id": i, "rtree": [i, i]} for i in range(10)]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def ft_index_and_items(db, namespace):
    """
    Create rtree index and items
    """
    db.index.create(namespace, {"name": "ft", "json_paths": ["ft"], "field_type": "string", "index_type": "text"})
    content = ["one word", "sword two", "three work 333"]
    items = [{"id": i, "ft": c} for i, c in enumerate(content)]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def array_index_and_items(db, namespace):
    """
    Create rtree index and items
    """
    db.index.create(namespace, {"name": "arr", "json_paths": ["arr"], "field_type": "int", "index_type": "tree",
                                "is_array": True})
    items = [{"id": i, "arr": [i, i + 1]} for i in range(5)]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def index_and_duplicate_items(db, namespace):
    """
    Create index and items with duplicate value
    """
    db.index.create(namespace, {"name": "idx", "json_paths": ["idx"], "field_type": "int", "index_type": "hash"})
    items = [{"id": 0, "idx": 0}, {"id": 1, "idx": 1}, {"id": 2, "idx": 1}, {"id": 3, "idx": 3}]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def array_indexes_and_items(db, namespace):
    """
    Create array indexes and items
    """
    db.index.create(namespace, {"name": "arr1", "json_paths": ["arr1"], "field_type": "int",
                                "index_type": "tree", "is_array": True})
    db.index.create(namespace, {"name": "arr2", "json_paths": ["arr2"], "field_type": "int",
                                "index_type": "tree", "is_array": True})
    items = [{"id": i, "arr1": [i, i % 2], "arr2": [i % 2, i]} for i in range(5)]
    for item in items:
        db.item.insert(namespace, item)
    yield items


@pytest.fixture
def metadata(db, namespace):
    """
    Put metadata  to namespace
    """
    key, value = 'key', 'value'
    db.meta.put(namespace, key, value)
    yield key, value


@pytest.fixture
def second_namespace(db):
    second_namespace_name = "test_ns_for_join"
    db.namespace.open(second_namespace_name)
    db.index.create(second_namespace_name, index_definition)
    yield second_namespace_name
    db.namespace.drop(second_namespace_name)


@pytest.fixture
def second_item(db, second_namespace):
    """
    Create item for the second namespace
    """
    item = {"id": 1, "second_ns_val": "second_ns_testval_1"}
    db.item.insert(second_namespace, item)
    yield item


@pytest.fixture
def second_items(db, second_namespace):
    """
    Create more items for the second namespace
    """
    items = [{"id": i, "second_ns_val": f"second_ns_testval_{i}"} for i in range(1, 6)]
    for item in items:
        db.item.insert(second_namespace, item)
    yield items



def insert_item_transaction(db, namespace, item_definition):
    """
    Insert an item into namespace using transaction
    """
    transaction = db.tx.begin(namespace)
    transaction.insert_item(item_definition)
    transaction.commit()


def upsert_item_transaction(db, namespace, item_definition):
    """
    Upsert or update an item into namespace using transaction
    """
    transaction = db.tx.begin(namespace)
    transaction.upsert_item(item_definition)
    transaction.commit()


def update_item_transaction(db, namespace, item_definition):
    """
    Update an item in namespace using transaction
    """
    transaction = db.tx.begin(namespace)
    transaction.update_item(item_definition)
    transaction.commit()


def delete_item_transaction(db, namespace, item_definition):
    """
    Delete item from namespace using transaction
    """
    transaction = db.tx.begin(namespace)
    transaction.delete_item(item_definition)
    transaction.commit()



from __future__ import annotations

from collections.abc import Iterable
from datetime import timedelta
from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from pyreindexer.exceptions import ApiError, QueryError
from pyreindexer.index_search_params import IndexSearchParamBruteForce, IndexSearchParamHnsw, IndexSearchParamIvf
from pyreindexer.point import Point
from pyreindexer.query_results import QueryResults


class ExtendedEnum(Enum):

    def __eq__(self, other):
        return self.value == other


class CondType(ExtendedEnum):
    CondAny = 0
    CondEq = 1
    CondLt = 2
    CondLe = 3
    CondGt = 4
    CondGe = 5
    CondRange = 6
    CondSet = 7
    CondAllSet = 8
    CondEmpty = 9
    CondLike = 10
    CondDWithin = 11


class StrictMode(ExtendedEnum):
    NotSet = 0
    Empty = 1
    Names = 2
    Indexes = 3


class JoinType(ExtendedEnum):
    LeftJoin = 0
    InnerJoin = 1
    OrInnerJoin = 2
    Merge = 3


class LogLevel(ExtendedEnum):
    Off = 0
    Error = 1
    Warning = 2
    Info = 3
    Trace = 4


simple_types = Union[int, str, bool, float]


class Query:
    """An object representing the context of a Reindexer query

    #### Attributes:
        api (module): An API module for Reindexer calls
        query_wrapper_ptr (int): A memory pointer to Reindexer query object
        err_code (int): The API error code
        err_msg (string): The API error message
        root (:object: Optional[`Query`]): The root query of the Reindexer query
        join_queries (list[:object:`Query`]): The list of join Reindexer query objects
        merged_queries (list[:object:`Query`]): The list of merged Reindexer query objects

    """

    def __init__(self, api, query_wrapper_ptr: int):
        """Constructs a new Reindexer query object

        #### Arguments:
            api (module): An API module for Reindexer calls
            query_wrapper_ptr (int): A memory pointer to Reindexer query object

        """

        self.api = api
        self.query_wrapper_ptr: int = query_wrapper_ptr
        self.err_code: int = 0
        self.err_msg: str = ''
        self.root: Optional[Query] = None
        self.join_queries: list[Query] = []
        self.merged_queries: list[Query] = []

    def __del__(self):
        """Frees query memory

        """

        if self.query_wrapper_ptr > 0:
            self.api.destroy_query(self.query_wrapper_ptr)

    def __raise_on_error(self) -> None:
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.err_code:
            raise ApiError(self.err_msg)

    @staticmethod
    def __where_knn_param(param: Union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]):
        """Validate and convert KNN search parameters

        #### Arguments:
            param (:obj:`Union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]`):
                KNN search parameters

        #### Returns:
            (:obj:`tuple[simple_types]`): actual parameters

        #### Raises:
            QueryError: Raises with an error message if no param are specified or have an invalid value

        """

        if param is None:
            raise QueryError("A required parameter is not specified. `param` can't be None")

        k: int = 0
        is_k: bool = False
        radius: float = 0.0
        is_radius: bool = False
        ef: int = 0
        nprobe: int = 0
        if isinstance(param, IndexSearchParamBruteForce):
            if param.k is not None:
                k = param.k
                is_k = True
            if param.radius is not None:
                radius = param.radius
                is_radius = True
        elif isinstance(param, IndexSearchParamHnsw):
            if param.k is not None and param.ef < param.k:
                raise QueryError("Ef should not be less than K")
            if param.k is not None:
                k = param.k
                is_k = True
            if param.radius is not None:
                radius = param.radius
                is_radius = True
            ef = param.ef
        elif isinstance(param, IndexSearchParamIvf):
            if param.nprobe < 1:
                raise QueryError("Nprobe should not be less than 1")
            if param.k is not None:
                k = param.k
                is_k = True
            if param.radius is not None:
                radius = param.radius
                is_radius = True
            nprobe = param.nprobe
        else:
            raise QueryError("Unexpected parameter type. `param` must be IndexSearchParamBruteForce, "
                             "IndexSearchParamHnsw or IndexSearchParamIvf")

        return k, is_k, radius, is_radius, ef, nprobe

    @staticmethod
    def __convert_to_list(param: Union[simple_types, tuple[list[simple_types], ...]]) -> list:
        """Converts an input parameter to a list of lists

        #### Arguments:
            param (union[simple_types, (list[simple_types], ...)]): The input parameter

        #### Returns:
            list[union[union[int, bool, float, str], list[union[int, bool, float, str]]]:
                Always converted to a list of lists

        """

        if param is None:
            return []

        if isinstance(param, str) or not isinstance(param, Iterable):
            return [param]

        if isinstance(param, list):
            return param

        result = list(param)
        if len(result) == 0 or not isinstance(result[0], list):
            result: list = [result]
        return result

    @staticmethod
    def __convert_strs_to_list(param: tuple[str, ...]) -> list[str]:
        """Converts an input parameter to a list

        #### Arguments:
            param (tuple[string, ...]): The input parameter

        #### Returns:
            list[string]: Always converted to a list

        """

        return [] if param is None else list(param)

    def __where(self, index: str, condition: CondType,
                keys: Union[simple_types, tuple[list[simple_types], ...]]) -> Query:
        """Adds where condition to DB query with args

        #### Arguments:
            index (string): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            keys (union[simple_types, (list[simple_types], ...)]):
                Value of index to be compared with. For composite indexes keys must be list,
                with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            QueryError: Raises with an error message if inappropriate condition is used
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if condition == CondType.CondDWithin:
            raise QueryError("In this case, use a special method 'dwithin'")

        params = self.__convert_to_list(keys)

        self.err_code, self.err_msg = self.api.where(self.query_wrapper_ptr, index, condition.value, params)
        self.__raise_on_error()
        return self

    def where(self, index: str, condition: CondType,
              keys: Union[simple_types, tuple[list[simple_types], ...]] = None) -> Query:
        """Adds where condition to DB query with args

        #### Arguments:
            index (string): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            keys (union[simple_types, (list[simple_types], ...)]):
                Value of index to be compared with. For composite indexes keys must be list,
                with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        return self.__where(index, condition, keys)

    def where_query(self, sub_query: Query, condition: CondType,
                    keys: Union[simple_types, tuple[list[simple_types], ...]] = None) -> Query:
        """Adds sub-query where condition to DB query with args

        #### Arguments:
            sub_query (:obj:`Query`): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            keys (union[simple_types, (list[simple_types], ...)]):
                Value of index to be compared with. For composite indexes keys must be list,
                with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        params = self.__convert_to_list(keys)

        self.err_code, self.err_msg = self.api.where_subquery(self.query_wrapper_ptr, sub_query.query_wrapper_ptr,
                                                              condition.value, params)
        self.__raise_on_error()
        return self

    def where_subquery(self, index: str, condition: CondType, sub_query: Query) -> Query:
        """Adds sub-query where condition to DB query

        #### Arguments:
            index (string): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            sub_query (:obj:`Query`): Field name used in condition clause

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.where_field_subquery(self.query_wrapper_ptr, index, condition.value, sub_query.query_wrapper_ptr)
        return self

    def where_composite(self, index: str, condition: CondType, keys: tuple[list[simple_types], ...]) -> Query:
        """Adds where condition to DB query with interface args for composite indexes

        #### Arguments:
            index (string): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            keys (list[simple_types], ...): Values of composite index to be compared with (value of each sub-index).
                Supported variants:
                    ([1, "test1"], [2, "test2"])
                    [[1, "test1"], [2, "test2"]])
                    ([1, "testval1"], )
                    [[1, "testval1"]]
                    (1, "testval1")

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        return self.__where(index, condition, keys)

    def where_uuid(self, index: str, condition: CondType, *uuids: UUID) -> Query:
        """Adds where condition to DB query with UUID.
            `index` MUST be declared as uuid-string index in this case

        #### Arguments:
            index (string): Field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            uuids (*:obj:`UUID`): Value of index to be compared with. For composite indexes uuids must be list,
                with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        params: list[str] = []
        for item in uuids:
            params.append(str(item))

        self.err_code, self.err_msg = self.api.where_uuid(self.query_wrapper_ptr, index, condition.value, params)
        self.__raise_on_error()
        return self

    def where_between_fields(self, first_field: str, condition: CondType, second_field: str) -> Query:
        """Adds comparing two fields where condition to DB query

        #### Arguments:
            first_field (string): First field name used in condition clause
            condition (:enum:`CondType`): Type of condition
            second_field (string): Second field name used in condition clause

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.where_between_fields(self.query_wrapper_ptr, first_field, condition.value, second_field)
        return self

    def where_knn(self, index: str, vec: List[float],
                  param: Union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]) -> Query:
        """Adds where condition to DB query with float_vector as args.
            `index` MUST be declared as float_vector index in this case

        #### Arguments:
            index (string): Field name used in condition clause (only float_vector)
            vec (list[float]): KNN value of index to be compared with
            param (:obj:`union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]`):
                KNN search parameters

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            QueryError: Raises with an error message if no vec are specified
            QueryError: Raises with an error message if no param are specified or have an invalid value
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if vec is None or len(vec) == 0:
            raise QueryError("A required parameter is not specified. `vec` can't be None or empty")

        k, is_k, radius, is_radius, ef, nprobe = self.__where_knn_param(param)

        self.err_code, self.err_msg =\
            self.api.where_knn(self.query_wrapper_ptr, index, is_k, k, is_radius, radius, ef, nprobe, vec)
        self.__raise_on_error()
        return self

    def where_knn_string(self, index: str, value: str,
                         param: Union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]) -> Query:
        """Adds where condition to DB query with string as args.
            `index` MUST be declared as float_vector index in this case.
            WARNING: Only relevant if automatic embedding is configured for this float_vector index

        #### Arguments:
            index (string): Field name used in condition clause (only float_vector)
            value (string): value to be generated using automatic embedding of KNN index value to be compared to
            param (:obj:`union[IndexSearchParamBruteForce|IndexSearchParamHnsw|IndexSearchParamIvf]`):
                KNN search parameters

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            QueryError: Raises with an error message if no value are specified
            QueryError: Raises with an error message if no param are specified or have an invalid value
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if len(value) == 0:
            raise QueryError("A required parameter is not specified. `value` can't be None or empty")

        k, is_k, radius, is_radius, ef, nprobe = self.__where_knn_param(param)

        self.err_code, self.err_msg =\
            self.api.where_knn_string(self.query_wrapper_ptr, index, is_k, k, is_radius, radius, ef, nprobe, value)
        self.__raise_on_error()
        return self

    def open_bracket(self) -> Query:
        """Opens bracket for where condition to DB query

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.open_bracket(self.query_wrapper_ptr)
        self.__raise_on_error()
        return self

    def close_bracket(self) -> Query:
        """Closes bracket for where condition to DB query

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.close_bracket(self.query_wrapper_ptr)
        self.__raise_on_error()
        return self

    def match(self, index: str, *keys: str) -> Query:
        """Adds string EQ-condition to DB query with string args

        #### Arguments:
            index (string): Field name used in condition clause
            keys (*string): Value of index to be compared with. For composite indexes keys must be list,
                with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        params: list = self.__convert_strs_to_list(keys)

        self.err_code, self.err_msg = self.api.where(self.query_wrapper_ptr, index, CondType.CondEq.value, params)
        self.__raise_on_error()
        return self

    def dwithin(self, index: str, point: Point, distance: float) -> Query:
        """Adds DWithin condition to DB query

        #### Arguments:
            index (string): Field name used in condition clause
            point (:obj:`Point`): Point object used in condition clause
            distance (float): Distance in meters between point

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.dwithin(self.query_wrapper_ptr, index, point.x, point.y, distance)
        return self

    def distinct(self, index: str) -> Query:
        """Performs distinct for a certain index. Return only items with uniq value of field

        #### Arguments:
            index (string): Field name for distinct operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.aggregate_distinct(self.query_wrapper_ptr, index)
        return self

    def aggregate_sum(self, index: str) -> Query:
        """Performs a summation of values for a specified index

        #### Arguments:
            index (string): Field name for sum operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.aggregate_sum(self.query_wrapper_ptr, index)
        return self

    def aggregate_avg(self, index: str) -> Query:
        """Finds for the average at the specified index

        #### Arguments:
            index (string): Field name for sum operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.aggregate_avg(self.query_wrapper_ptr, index)
        return self

    def aggregate_min(self, index: str) -> Query:
        """Finds for the minimum at the specified index

        #### Arguments:
            index (string): Field name for sum operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.aggregate_min(self.query_wrapper_ptr, index)
        return self

    def aggregate_max(self, index: str) -> Query:
        """Finds for the maximum at the specified index

        #### Arguments:
            index (string): Field name for sum operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.aggregate_max(self.query_wrapper_ptr, index)
        return self

    class _AggregateFacet:
        """An object representing the context of a Reindexer aggregate facet

        #### Attributes:
            api (module): An API module for Reindexer calls
            query_wrapper_ptr (int): A memory pointer to Reindexer query object

        """

        def __init__(self, query: Query):
            """Constructs a new Reindexer AggregateFacetRequest object

            #### Arguments:
                (:obj:`Query`): Query object for further customizations

            """

            self.api = query.api
            self.query_wrapper_ptr = query.query_wrapper_ptr

        def limit(self, limit: int) -> Query._AggregateFacet:
            """Limits facet aggregation results

            #### Arguments:
                limit (int): Limit of aggregation of facet

            #### Returns:
                (:obj:`_AggregateFacet`): Facet object for further customizations

            """

            self.api.aggregation_limit(self.query_wrapper_ptr, limit)
            return self

        def offset(self, offset: int) -> Query._AggregateFacet:
            """Sets offset of the facet aggregation results

            #### Arguments:
                limit (int): Offset in facet aggregation results

            #### Returns:
                (:obj:`_AggregateFacet`): Facet object for further customizations

            """

            self.api.aggregation_offset(self.query_wrapper_ptr, offset)
            return self

        def sort(self, field: str, desc: bool = False) -> Query._AggregateFacet:
            """Sorts facets by field value

            #### Arguments:
                field (str): Item field. Use field `count` to sort by facet's count value
                desc (bool): Sort in descending order flag

            #### Returns:
                (:obj:`_AggregateFacet`): Facet object for further customizations

            """

            self.api.aggregation_sort(self.query_wrapper_ptr, field, desc)
            return self

    def aggregate_facet(self, *fields: str) -> Query._AggregateFacet:
        """Gets fields facet value. Applicable to multiple data fields and the result of that could be sorted
            by any data column or `count` and cut off by offset and limit. In order to support this functionality
            this method returns AggregationFacetRequest which has methods sort, limit and offset

        #### Arguments:
            fields (*string): Fields any data column name or `count`, fields should not be empty

        #### Returns:
            (:obj:`_AggregateFacet`): Request object for further customizations

        """

        params: list = self.__convert_strs_to_list(fields)

        self.err_code, self.err_msg = self.api.aggregation(self.query_wrapper_ptr, params)
        self.__raise_on_error()
        return self._AggregateFacet(self)

    def sort(self, index: str, desc: bool = False,
             forced_sort_values: Union[simple_types, tuple[list[simple_types], ...]] = None) -> Query:
        """Applies sort order to return from query items. If forced_sort_values argument specified, then items equal to
            values, if found will be placed in the top positions. Forced sort is support for the first sorting field
            only

        #### Arguments:
            index (string): The index name
            desc (bool): Sort in descending order
            forced_sort_values (union[simple_types, (list[simple_types], ...)]):
                Value of index to match. For composite indexes keys must be list, with value of each sub-index

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        values = self.__convert_to_list(forced_sort_values)

        self.err_code, self.err_msg = self.api.sort(self.query_wrapper_ptr, index, desc, values)
        self.__raise_on_error()
        return self

    def sort_stpoint_distance(self, index: str, point: Point, desc: bool) -> Query:
        """Applies geometry sort order to return from query items. Wrapper for geometry sorting by shortest distance
            between geometry field and point (ST_Distance)

        #### Arguments:
            index (string): The index name
            point (:obj:`Point`): Point object used in sorting operation
            desc (bool): Sort in descending order

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        request: str = f"ST_Distance({index},ST_GeomFromText('point({point.x:.12f} {point.y:.12f})'))"
        return self.sort(request, desc)

    def sort_stfield_distance(self, first_field: str, second_field: str, desc: bool) -> Query:
        """Applies geometry sort order to return from query items. Wrapper for geometry sorting by shortest distance
            between 2 geometry fields (ST_Distance)

        #### Arguments:
            first_field (string): First field name used in condition
            second_field (string): Second field name used in condition
            desc (bool): Sort in descending order

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        request: str = f"ST_Distance({first_field},{second_field})"
        return self.sort(request, desc)

    def op_and(self) -> Query:
        """Next condition will be added with AND.
            This is the default operation for WHERE statement. Do not have to be called explicitly in user's code.
            Used in DSL conversion

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.op_and(self.query_wrapper_ptr)
        return self

    def op_or(self) -> Query:
        """Next condition will be added with OR.
            Implements short-circuiting:
            if the previous condition is successful the next will not be evaluated, but except Join conditions

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.op_or(self.query_wrapper_ptr)
        return self

    def op_not(self) -> Query:
        """Next condition will be added with NOT AND.
            Implements short-circuiting: if the previous condition is failed the next will not be evaluated

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.op_not(self.query_wrapper_ptr)
        return self

    def request_total(self) -> Query:
        """Requests total items calculation

        #### Arguments:
            total_name (string, optional): Name to be requested

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.request_total(self.query_wrapper_ptr)
        return self

    def cached_total(self) -> Query:
        """Requests cached total items calculation

        #### Arguments:
            total_name (string, optional): Name to be requested

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.cached_total(self.query_wrapper_ptr)
        return self

    def limit(self, limit_items: int) -> Query:
        """Sets a limit (count) of returned items. Analog to sql LIMIT rowsNumber

        #### Arguments:
            limit_items (int): Number of rows to get from result set

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.limit(self.query_wrapper_ptr, limit_items)
        return self

    def offset(self, start_offset: int) -> Query:
        """Sets the number of the first selected row from result query

        #### Arguments:
            limit_items (int): Index of the first row to get from result set

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.offset(self.query_wrapper_ptr, start_offset)
        return self

    def debug(self, level: LogLevel) -> Query:
        """Changes debug log level on server

        #### Arguments:
            level (:enum:`LogLevel`): Debug log level on server

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.debug(self.query_wrapper_ptr, level.value)
        return self

    def strict(self, mode: StrictMode) -> Query:
        """Changes strict mode

        #### Arguments:
            mode (:enum:`StrictMode`): Strict mode

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.strict(self.query_wrapper_ptr, mode.value)
        return self

    def explain(self) -> Query:
        """Enables explain query

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.explain(self.query_wrapper_ptr)
        return self

    def with_rank(self) -> Query:
        """Outputs fulltext/float_vector rank. Allowed only with fulltext and KNN query

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.with_rank(self.query_wrapper_ptr)
        return self

    def execute(self, timeout: timedelta = timedelta(milliseconds=0)) -> QueryResults:
        """Executes a select query

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum 1 millisecond, if set to a value less, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`QueryResults`): A QueryResults iterator

        #### Raises:
            ApiError: Raises with an error message when query is in an invalid state
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.root is not None:
            return self.root.execute(timeout)

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        (self.err_code, self.err_msg,
         wrapper_ptr, iter_count, total_count) = self.api.select_query(self.query_wrapper_ptr, milliseconds)
        self.__raise_on_error()
        return QueryResults(self.api, wrapper_ptr, iter_count, total_count)

    def delete(self, timeout: timedelta = timedelta(milliseconds=0)) -> int:
        """Executes a query, and delete items, matches query

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum 1 millisecond, if set to a value less, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (int): Number of deleted elements

        #### Raises:
            QueryError: Raises with an error message when query is in an invalid state
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if (self.root is not None) or (len(self.join_queries) > 0):
            raise QueryError("Delete does not support joined queries")

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        self.err_code, self.err_msg, number = self.api.delete_query(self.query_wrapper_ptr, milliseconds)
        self.__raise_on_error()
        return number

    def set_object(self, field: str, values: list[simple_types]) -> Query:
        """Adds an update query to an object field for an update query

        #### Arguments:
            field (string): Field name
            values (list[simple_types]): List of values to add

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            QueryError: Raises with an error message if no values are specified
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if values is None:
            raise QueryError("A required parameter is not specified. `values` can't be None")

        self.err_code, self.err_msg = self.api.set_object(self.query_wrapper_ptr, field, values)
        self.__raise_on_error()
        return self

    def set(self, field: str, values: list[simple_types]) -> Query:
        """Adds a field update request to the update request

        #### Arguments:
            field (string): Field name
            values (list[simple_types]): List of values to add

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        values = [] if values is None else values

        self.err_code, self.err_msg = self.api.set(self.query_wrapper_ptr, field, values)
        self.__raise_on_error()
        return self

    def drop(self, index: str) -> Query:
        """Drops a value for a field

        #### Arguments:
            index (string): Field name for drop operation

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.drop(self.query_wrapper_ptr, index)
        return self

    def expression(self, field: str, value: str) -> Query:
        """Updates indexed field by arithmetical expression

        #### Arguments:
            field (string): Field name
            value (string): New value expression for field

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        self.api.expression(self.query_wrapper_ptr, field, value)
        return self

    def update(self, timeout: timedelta = timedelta(milliseconds=0)) -> QueryResults:
        """Executes update query, and update fields in items, which matches query

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum 1 millisecond, if set to a value less, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`QueryResults`): A QueryResults iterator

        #### Raises:
            QueryError: Raises with an error message when query is in an invalid state
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if (self.root is not None) or (len(self.join_queries) > 0):
            raise QueryError("Update does not support joined queries")

        milliseconds: int = int(timeout / timedelta(milliseconds=1))
        (self.err_code, self.err_msg,
         wrapper_ptr, iter_count, total_count) = self.api.update_query(self.query_wrapper_ptr, milliseconds)
        self.__raise_on_error()
        return QueryResults(self.api, wrapper_ptr, iter_count, total_count)

    def must_execute(self, timeout: timedelta = timedelta(milliseconds=0)) -> QueryResults:
        """Executes a query, and update fields in items, which matches query, with status check

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum 1 millisecond, if set to a value less, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:obj:`QueryResults`): A QueryResults iterator

        #### Raises:
            ApiError: Raises with an error message when query is in an invalid state
            ApiError: Raises with an error message of API return on non-zero error code

        """

        result = self.execute(timeout)
        result.status()
        return result

    def get(self, timeout: timedelta = timedelta(milliseconds=0)) -> (str, bool):
        """Executes a query, and return 1 JSON item

        #### Arguments:
            timeout (`datetime.timedelta`): Optional timeout for performing a server-side operation.
                Minimum 1 millisecond, if set to a value less, it corresponds to disabling the timeout.
                A value of 0 disables the timeout (default value)

        #### Returns:
            (:tuple:string,bool): 1st string item and found flag

        #### Raises:
            ApiError: Raises with an error message when query is in an invalid state
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.root is not None:
            return self.get()

        selected_items = self.limit(1).must_execute(timeout)
        for item in selected_items:
            return item, True

        return '', False

    def __join(self, query: Query, field: str, join_type: JoinType) -> Query:
        """Joins queries

        #### Arguments:
            query (:obj:`Query`): Query object to join
            field (string): Joined field name
            type (:enum:`JoinType`): Join type

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            Exception: Raises with an error message when query is in an invalid state

        """

        if self.root is not None:
            return self.root.__join(query, field, join_type)

        if query.root is not None:
            raise QueryError("Query.join call on already joined query. You should create new Query")

        # index of join query
        self.api.join(self.query_wrapper_ptr, join_type.value, query.query_wrapper_ptr)

        query.root = self
        self.join_queries.append(query)
        return query

    def inner_join(self, query: Query, field: str) -> Query:
        """Joins 2 queries.
            Items from the 1-st query are filtered by and expanded with the data from the 2-nd query

        #### Arguments:
            query (:obj:`Query`): Query object to left join
            field (string): Joined field name. As unique identifier for the join between this query and `join_query`.
                Parameter in order for InnerJoin to work: namespace of `query` contains `field` as one of its fields
                marked as `joined`

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        return self.__join(query, field, JoinType.InnerJoin)

    def join(self, query: Query, field: str) -> Query:
        """Join is an alias for LeftJoin. Joins 2 queries.
            Items from this query are expanded with the data from the `query`

        #### Arguments:
            query (:obj:`Query`): Query object to left join
            field (string): Joined field name. As unique identifier for the join between this query and `join_query`

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        return self.__join(query, field, JoinType.LeftJoin)

    def left_join(self, join_query: Query, field: str) -> Query:
        """Joins 2 queries.
            Items from this query are expanded with the data from the join_query.
            One of the conditions below must hold for `field` parameter in order for LeftJoin to work:
                namespace of `join_query` contains `field` as one of its fields marked as `joined`

        #### Arguments:
            query (:obj:`Query`): Query object to left join
            field (string): Joined field name. As unique identifier for the join between this query and `join_query`

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        return self.__join(join_query, field, JoinType.LeftJoin)

    def merge(self, query: Query) -> Query:
        """Merges queries of the same type

        #### Arguments:
            query (:obj:`Query`): Query object to merge

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        """

        if self.root is not None:
            return self.root.merge(query)

        if query.root is not None:
            query = query.root

        query.root = self
        self.merged_queries.append(query)
        self.api.merge(self.query_wrapper_ptr, query.query_wrapper_ptr)
        return self

    def on(self, index: str, condition: CondType, join_index: str) -> Query:
        """On specifies join condition

        #### Arguments:
            index (string): Field name from `Query` namespace should be used during join
            condition (:enum:`CondType`): Type of condition, specifies how `Query` will be joined with the latest join query issued on `Query` (e.g. `EQ`/`GT`/`SET`/...)
            join_index (string): Index-field name from namespace for the latest join query issued on `Query` should be used during join

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            QueryError: Raises with an error message when query is in an invalid state

        """

        if self.root is None:
            raise QueryError("Can't join on root query")

        self.api.on(self.query_wrapper_ptr, index, condition.value, join_index)
        return self

    def select(self, *fields: str) -> Query:
        """Sets list of columns in this namespace to be finally selected.
            The columns should be specified in the same case as the jsonpaths corresponding to them.
            Non-existent fields and fields in the wrong case are ignored.
            If there are no fields in this list that meet these conditions, then the filter works as "*"

        #### Arguments:
            fields (*string): List of columns to be selected

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        keys: list = self.__convert_strs_to_list(fields)

        self.err_code, self.err_msg = self.api.select_filter(self.query_wrapper_ptr, keys)
        self.__raise_on_error()
        return self

    def functions(self, *functions: str) -> Query:
        """Adds sql-functions to query

        #### Arguments:
            functions (*string): Functions declaration

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        funcs: list = self.__convert_strs_to_list(functions)

        self.err_code, self.err_msg = self.api.functions(self.query_wrapper_ptr, funcs)
        self.__raise_on_error()
        return self

    def equal_position(self, *equal_position: str) -> Query:
        """Adds equal position fields to arrays queries

        #### Arguments:
            equal_poses (*string): Equal position fields to arrays queries

        #### Returns:
            (:obj:`Query`): Query object for further customizations

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        equal_pos: list = self.__convert_strs_to_list(equal_position)

        self.err_code, self.err_msg = self.api.equal_position(self.query_wrapper_ptr, equal_pos)
        self.__raise_on_error()
        return self



from pyreindexer.exceptions import ApiError


class QueryResults:
    """QueryResults is a disposable iterator of Reindexer results for such queries as SELECT etc.
        When the results are fetched the iterator closes and frees a memory of results buffer of Reindexer

    #### Attributes:
        api (module): An API module for Reindexer calls
        err_code (int): The API error code
        err_msg (string): The API error message
        qres_wrapper_ptr (int): A memory pointer to Reindexer iterator object
        qres_iter_count (int): A count of results for iterations
        pos (int): The current result position in iterator

    """

    def __init__(self, api, qres_wrapper_ptr, qres_iter_count, qres_total_count):
        """Constructs a new Reindexer query results iterator object

        #### Arguments:
            api (module): An API module for Reindexer calls
            qres_wrapper_ptr (int): A memory pointer to Reindexer iterator object
            qres_iter_count (int): A count of results for iterations
            qres_total_count (int): A total or cached count of results

        """

        self.api = api
        self.qres_wrapper_ptr = qres_wrapper_ptr
        self.qres_iter_count = qres_iter_count
        self.qres_total_count = qres_total_count
        self.pos = 0
        self.err_code = 0
        self.err_msg = ""

    def __raise_on_error(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.err_code:
            raise ApiError(self.err_msg)

    def __iter__(self):
        """Returns the current iteration result

        """

        return self

    def __next__(self):
        """Returns the next iteration result

        #### Raises:
            StopIteration: Frees results on end of iterator and raises with iteration stop

        """

        if self.pos < self.qres_iter_count:
            self.pos += 1
            self.err_code, self.err_msg, res = self.api.query_results_iterate(self.qres_wrapper_ptr)
            self.__raise_on_error()
            return res
        else:
            del self
            raise StopIteration

    def __del__(self):
        """Calls close iterator method on an iterator object deletion

        """

        self._close_iterator()

    def status(self) -> None:
        """Check status

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg = self.api.query_results_status(self.qres_wrapper_ptr)
        self.__raise_on_error()

    def count(self) -> int:
        """Returns a count of results for iterations

        #### Returns
            int: A count of results

        """

        return self.qres_iter_count

    def total_count(self) -> int:
        """Returns a total or cached count of results

        #### Returns
            int: A total or cached count of results

        """

        return self.qres_total_count

    def _close_iterator(self) -> None:
        """Frees query results for the current iterator

        """

        self.qres_iter_count = 0
        self.api.query_results_delete(self.qres_wrapper_ptr)

    def get_agg_results(self) -> dict:
        """Returns aggregation results for the current query

        #### Returns
            (:obj:`dict`): Dictionary with all results for the current query

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg, res = self.api.get_agg_results(self.qres_wrapper_ptr)
        self.__raise_on_error()
        return res

    def get_explain_results(self) -> str:
        """Returns explain results for the current query

        #### Returns
            (string): Formatted string with explain of results for the current query

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        self.err_code, self.err_msg, res = self.api.get_explain_results(self.qres_wrapper_ptr)
        self.__raise_on_error()
        return res



import os
import shlex
import subprocess

import requests
from requests.adapters import HTTPAdapter, Retry

from tests.helpers.log_helper import log_fixture


session = requests.Session()
Retry.DEFAULT_BACKOFF_MAX = 0.03
retry = Retry(total=100, backoff_factor=0.03)
adapter = HTTPAdapter(max_retries=retry)
session.mount("http://", adapter)


class ReindexerServer:

    def __init__(self, rx_bin_path="", http_port=9088, rpc_port=6534, storage="/tmp/reindex_test",
                 auth=False, user=None, password=None):
        self.rx_bin_path = rx_bin_path
        self.http_port = http_port
        self.rpc_port = rpc_port
        self.httpaddr = f"127.0.0.1:{http_port}"
        self.rpcaddr = f"127.0.0.1:{rpc_port}"

        self.storage = storage

        self.auth = auth
        self.user = user
        self.password = password

        self.proc = None

        self.__cmd = f"{os.path.join(self.rx_bin_path, 'reindexer_server')} --db {self.storage} --httpaddr={self.httpaddr} --rpcaddr={self.rpcaddr}"
        if self.auth:
            self.__cmd += " --security"

    def run(self, *args):
        command = shlex.split(self.__cmd) + list(args)
        self.proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     universal_newlines=True, bufsize=1)
        self.is_running()
        log_fixture.info(f"Reindexer server started at {self.http_port} http port")

    def is_running(self):
        try:
            url = f"http://{self.httpaddr}/api/v1/check"
            auth = (self.user, self.password) if self.auth else None
            session.get(url=url, auth=auth)
            return True
        except (ConnectionError, requests.exceptions.ConnectionError, RecursionError, ChildProcessError,
                AssertionError):
            self.terminate()
            if not hasattr(self.proc, "returncode"):
                raise ChildProcessError("Server is not running")
            else:
                raise ChildProcessError(f"Server status code is {self.proc.returncode}")

    def terminate(self):
        try:
            self.proc.terminate()
            self.proc.wait(timeout=1)
        except subprocess.TimeoutExpired:
            self.proc.kill()



class IndexSearchParamBruteForce:
    """Index search param for brute force index. Equal to basic parameters

    #### Attributes:
        k (int): Expected size of KNN index results. Should not be less than 1
        radius (float): In addition to the parameter `k`, the query results can also be filtered by a `rank` -
            value using the parameter, witch called `radius`. It's named so because, under the `L2`-metric,
            it restricts vectors from query result to a sphere of the specified radius

    """

    def __init__(self, k: int = None, radius: float = None):
        if k is None and radius is None:
            raise ValueError("Either 'k' or 'radius' needs to be specified")
        if k is not None and k < 1:
            raise ValueError("KNN limit 'k' should not be less than 1")
        self.k = k
        self.radius = radius


class IndexSearchParamHnsw:
    """Index search param for HNSW index.

    #### Attributes:
        k (int): Expected size of KNN index results. Should not be less than 1
        ef (int): Size of nearest neighbor buffer that will be filled during fetching. Should not be less than 'k',
            good story when `ef` ~= 1.5 * `k`
        radius (float): In addition to the parameter `k`, the query results can also be filtered by a `rank` -
            value using the parameter, witch called `radius`. It's named so because, under the `L2`-metric,
            it restricts vectors from query result to a sphere of the specified radius

    """

    def __init__(self, ef: int, k: int = None, radius: float = None):
        if k is None and radius is None:
            raise ValueError("Either 'k' or 'radius' needs to be specified")
        if k is not None:
            if k < 1:
                raise ValueError("KNN limit 'k' should not be less than 1")
            if ef < k:
                raise ValueError("'ef' should not be less than 'k'")
        self.k = k
        self.radius = radius
        self.ef = ef


class IndexSearchParamIvf:
    """Index search param for IVF index.

    #### Attributes:
        k (int): Expected size of KNN index results. Should not be less than 1
        nprobe (int): Number of centroids that will be scanned in where. Should not be less than 1
        radius (float): In addition to the parameter `k`, the query results can also be filtered by a `rank` -
            value using the parameter, witch called `radius`. It's named so because, under the `L2`-metric,
            it restricts vectors from query result to a sphere of the specified radius

    """

    def __init__(self, nprobe: int, k: int = None, radius: float = None):
        if k is None and radius is None:
            raise ValueError("Either 'k' or 'radius' needs to be specified")
        if k is not None and k < 1:
            raise ValueError("KNN limit 'k' should not be less than 1")
        if nprobe < 1:
            raise ValueError("'nprobe' should not be less than 1")
        self.nprobe = nprobe
        self.k = k
        self.radius = radius



import copy
from datetime import timedelta
from typing import Final

from hamcrest import *

from pyreindexer.exceptions import ApiError
from tests.helpers.base_helper import random_vector
from tests.helpers.check_helper import check_response_has_close_to_ns_items
from tests.test_data.constants import vector_index_bf, vector_index_hnsw, vector_index_ivf


class TestSqlQueries:
    def test_sql_select(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Execute SQL query SELECT")
        query = f"SELECT * FROM {namespace}"
        items_list = list(db.query.sql(query, timeout=timedelta(milliseconds=1000)))
        # Then ("Check that selected item is in result")
        assert_that(items_list, equal_to([item]), "Can't SQL select data")

    def test_sql_select_with_join(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces")
        # When ("Execute SQL query SELECT with JOIN")
        query = f"SELECT id FROM {namespace} INNER JOIN {second_namespace} " \
                f"ON {namespace}.id = {second_namespace}.id"
        item_list = list(db.query.sql(query))
        # Then ("Check that selected item is in result")
        item_with_joined = {'id': 1, f'joined_{second_namespace}': [second_item]}
        assert_that(item_list, equal_to([item_with_joined]),
                    "Can't SQL select data with JOIN")

    def test_sql_select_with_condition(self, db, namespace, index, items):
        # Given("Create namespace with item")
        # When ("Execute SQL query SELECT")
        item_id = 3
        query = f"SELECT * FROM {namespace} WHERE id = {item_id}"
        items_list = list(db.query.sql(query))
        # Then ("Check that selected item is in result")
        assert_that(items_list, equal_to([items[item_id]]), "Can't SQL select data with condition")

    def test_sql_update(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Execute SQL query UPDATE")
        item["val"] = "new_val"
        query = f"UPDATE {namespace} SET \"val\" = '{item['val']}' WHERE id = 100"
        items_list = list(db.query.sql(query))
        # Then ("Check that item is updated")
        assert_that(items_list, equal_to([item]), "Can't SQL update data")

    def test_sql_delete(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Execute SQL query DELETE")
        query_delete = f"DELETE FROM {namespace} WHERE id = 100"
        db.query.sql(query_delete)
        # Then ("Check that item is deleted")
        query_select = f"SELECT * FROM {namespace}"
        item_list = list(db.query.sql(query_select))
        assert_that(item_list, empty(), "Can't SQL delete data")

    def test_sql_select_with_syntax_error(self, db, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Execute SQL query SELECT with incorrect syntax")
        query = "SELECT *"
        # Then ("Check that selected item is in result")
        assert_that(calling(db.query.sql).with_args(query),
                    raises(ApiError, pattern="Expected .* but found"),
                    "Error wasn't raised when syntax was incorrect")

    def test_sql_select_with_aggregations(self, db, namespace, index, items):
        # Given("Create namespace with item")
        # When ("Insert items into namespace")
        select_result = db.query.sql(f"SELECT min(id), max(id), avg(id), sum(id) FROM {namespace}").get_agg_results()
        assert_that(select_result, has_length(4), "The aggregation result does not contain all aggregations")

        ids = [i["id"] for i in items]
        expected_values = {"min": min(ids), "max": max(ids), "avg": sum(ids) / len(items), "sum": sum(ids)}

        # Then ("Check that returned agg results are correct")
        for agg in select_result:
            assert_that(agg['value'], equal_to(expected_values[agg['type']]),
                        f"Incorrect aggregation result for {agg['type']}")

    def test_sql_select_timeout_small(self, db, namespace, index):
        # Given("Create namespace with items")
        items = [{"id": i, "val": f"testval{i}", "non_idx": i + 1} for i in range(10000)]
        for item in items:
            db.item.insert("new_ns", item)
        # When ("Try to execute SQL query SELECT with small timeout")
        query = ("SELECT * FROM new_ns WHERE id > -1 AND non_idx > 0 MERGE "
                 "(SELECT * FROM new_ns WHERE val < 'testval1000' AND non_idx > 0) MERGE "
                 "(SELECT * FROM new_ns WHERE val > 'testval1000' AND id RANGE(1,9000) AND non_idx > 100) MERGE "
                 "(SELECT * FROM new_ns WHERE non_idx > id AND NOT (id < 100))")
        assert_that(calling(db.query.sql).with_args(query, timeout=timedelta(milliseconds=1)),
                    raises(ApiError, pattern="Context timeout|Read lock (.*) was canceled or timed out (mutex)"))


class TestSqlQueriesKNN:

    def test_sql_select_knn_brute_force(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_bf)
        index["config"] = {"dimension": dimension, "metric": "l2", "start_size": 1000}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        for item in items:
            db.item.insert("new_ns", item)
        # When ("Execute SQL query SELECT KNN")
        value = random_vector(dimension)
        k = 47
        query = f"SELECT *, vectors() FROM {namespace} WHERE KNN(vec, {value}, k={k})"
        query_result = list(db.query.sql(query, timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)
        assert_that(query_result, has_length(k))

    def test_sql_select_knn_hnsw(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_hnsw)
        index["config"] = {"dimension": dimension, "metric": "inner_product", "start_size": 1000, "m": 16,
                           "ef_construction": 200}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        for item in items:
            db.item.insert("new_ns", item)
        # When ("Execute SQL query SELECT KNN")
        value = random_vector(dimension)
        query = f"SELECT *, vectors() FROM {namespace} WHERE KNN(vec, {value}, k=20, ef=30)"
        query_result = list(db.query.sql(query, timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    def test_sql_select_knn_ivf(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_ivf)
        index["config"] = {"dimension": dimension, "metric": "l2", "centroids_count": 3}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(150)]
        for item in items:
            db.item.insert("new_ns", item)
        # When ("Execute SQL query SELECT KNN")
        value = random_vector(dimension)
        query = f"SELECT *, vectors() FROM {namespace} WHERE KNN(vec, {value}, k=30, nprobe=2)"
        query_result = list(db.query.sql(query, timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)



from pyreindexer.exceptions import ApiError


class RaiserMixin:
    """RaiserMixin contains methods for checking some typical API bad events and raise if there is a necessity

    """
    err_code: int
    err_msg: str
    rx: int

    def raise_on_error(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ApiError: Raises with an error message of API return on non-zero error code

        """

        if self.err_code:
            raise ApiError(self.err_msg)

    def raise_on_not_init(self):
        """Checks if there is an error code and raises with an error message

        #### Raises:
            ConnectionError: Raises with an error message when Reindexer instance is not initialized yet

        """

        if self.rx <= 0:
            raise ConnectionError("Connection is not initialized")


def raise_if_error(func):
    def wrapper(self, *args, **kwargs):
        self.raise_on_not_init()
        res = func(self, *args, **kwargs)
        self.raise_on_error()
        return res

    return wrapper



import copy
import json
import random
import uuid
from datetime import timedelta
from typing import Final

import pytest
from hamcrest import *

from pyreindexer.exceptions import ApiError, QueryError
from pyreindexer.index_search_params import IndexSearchParamBruteForce, IndexSearchParamHnsw, IndexSearchParamIvf
from pyreindexer.point import Point
from pyreindexer.query import CondType, LogLevel, StrictMode
from tests.helpers.base_helper import calculate_distance, get_ns_items, random_vector
from tests.helpers.check_helper import check_response_has_close_to_ns_items, check_response_has_only_close_to_items
from tests.test_data.constants import AGGREGATE_FUNCTIONS_MATH, vector_index_bf, vector_index_hnsw, vector_index_ivf


class TestQuerySelect:
    def test_query_select_where(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query_result = list(query.where("id", CondType.CondEq, 3).execute(timeout=timedelta(seconds=1)))
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[3]]), "Wrong query results")

    def test_query_select_where_array(self, db, namespace, index, array_index_and_items):
        # Given("Create namespace with array index and items")
        items = array_index_and_items
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query_result = list(query.where("arr", CondType.CondEq, [2, 3]).must_execute())
        # Then ("Check that selected item is in result")
        expected_items = [items[i] for i in [1, 2, 3]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_all(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query_result = list(query.execute())
        # Then ("Check that all items are in result")
        assert_that(query_result, equal_to(items), "Wrong query results")

    def test_query_select_fields(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with select fields")
        select_result = list(query.select("id").must_execute())
        # Then ("Check that selected items are in result")
        ids = [{"id": i["id"]} for i in items]
        assert_that(select_result, equal_to(ids), "Wrong query results")

    def test_query_select_get_true(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with get")
        query_result = list(query.get())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[0], True]), "Wrong query results")

    def test_query_select_get_false(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with get")
        query_result = list(query.where("id", CondType.CondEq, 99).get())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to(["", False]), "Wrong query results")

    def test_query_select_where_cond_any(self, db, namespace, index, sparse_index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with cond any")
        query_result = list(query.where("val", CondType.CondAny, None).must_execute())
        # Then ("Check that all items is in result")
        assert_that(query_result, equal_to(items), "Wrong query results")

    def test_query_select_where_cond_empty(self, db, namespace, index, sparse_index, items):
        # Given("Create namespace with index and items")
        # Given ("Create item without sparse field")
        item_empty = {"id": -1}
        db.item.insert(namespace, item_empty)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with cond empty")
        query_result = list(query.where("val", CondType.CondEmpty, None).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([item_empty]), "Wrong query results")

    def test_query_select_where_cond_like(self, db, namespace, indexes, items):
        # Given("Create namespace with indexes and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with cond like")
        query_result = list(query.where("val", CondType.CondLike, "test%").must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to(items), "Wrong query results")

    def test_query_select_where_query_field(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        sub_query = (db.query.new(namespace)
                     .select("id")
                     .where("id", CondType.CondLt, 5)
                     .where("id", CondType.CondGe, 2))
        # When ("Make select query with where_subquery")
        query_result = list(db.query.new(namespace).where_subquery('id', CondType.CondSet, sub_query).execute())
        # Then ("Check that selected item is in result")
        expected_items = [items[i] for i in [2, 3, 4]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_where_query_all(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        sub_query = db.query.new(namespace).select("id").where("id", CondType.CondLt, 5)
        # When ("Make select query with where_query and subquery")
        query_result = list(db.query.new(namespace).where_query(sub_query, CondType.CondGe, 2).execute())
        # Then ("Check that all items selected in result")
        assert_that(query_result, equal_to(items), "Wrong query results")

    def test_query_select_where_query_empty(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        sub_query = db.query.new(namespace).select("id").where("id", CondType.CondEq, 5)
        # When ("Make select query with where_query and subquery")
        query_result = list(db.query.new(namespace).where_query(sub_query, CondType.CondLe, 2).execute())
        # Then ("Check that result is empty")
        assert_that(query_result, empty(), "Wrong query results")

    def test_query_select_where_composite(self, db, namespace, composite_index, items):
        # Given("Create namespace with composite index")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with where_composite")
        query_result = list(query.where_composite("comp_idx", CondType.CondEq, (1, "testval1")).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1]]), "Wrong query results")

    @pytest.mark.parametrize("value", [
        [[1, "testval1"], [2, "testval2"]],
        [(1, "testval1"), (2, "testval2")],
        ([1, "testval1"], [2, "testval2"])
    ])
    def test_query_select_where_composite_multiple_results(self, db, namespace, composite_index, items, value):
        # Given("Create namespace with composite index")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with where_composite")
        query_result = list(query.where_composite("comp_idx", CondType.CondEq, value).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1], items[2]]), "Wrong query results")

    def test_query_select_where_uuid(self, db, namespace, index):
        # Given("Create namespace with index")
        # Given ("Create uuid index")
        db.index.create(namespace, {"name": "uuid", "json_paths": ["uuid"], "field_type": "uuid", "index_type": "hash"})
        # Given ("Create items")
        items = [{"id": 0, "uuid": "f0dbda48-1b92-4aa1-b4dc-39b6867a202e"},
                 {"id": 1, "uuid": "189005c5-106f-4582-9838-f7a5b18649fc"},
                 {"id": 2, "uuid": "50005725-dd9d-4891-a9bc-7c7cb0ed2a3d"}]
        for item in items:
            db.item.insert(namespace, item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with where_uuid")
        item = items[1]
        some_uuid = uuid.UUID(item["uuid"])
        query_result = list(query.where_uuid("uuid", CondType.CondEq, some_uuid).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([item]), "Wrong query results")

    def test_query_select_where_between_fields(self, db, namespace, index):
        # Given("Create namespace with index")
        # Given("Create items")
        items = [{"id": 0, "age": 0.5}, {"id": 1, "age": 1}, {"id": 2, "age": 20}]
        for item in items:
            db.item.insert(namespace, item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with where_between_fields")
        query_result = list(query.where_between_fields("id", CondType.CondEq, "age").must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1]]), "Wrong query results")

    def test_query_select_with_brackets_and_ops(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with brackets and different op")
        query = (query.where("id", CondType.CondEq, 0)
                 .op_or().open_bracket()
                 .op_and().where("id", CondType.CondGe, 2)
                 .op_or().where("id", CondType.CondLt, 7)
                 .op_not().where("id", CondType.CondSet, [1, 6])

                 .op_and().open_bracket()
                 .op_not().where("id", CondType.CondRange, [5, 9])
                 .close_bracket()

                 .close_bracket())
        query_result = list(query.must_execute())
        # Then ("Check that selected items are in result")
        expected_items = [items[i] for i in [0, 2, 3, 4]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_match(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create string index")
        db.index.create(namespace, {"name": "val", "json_paths": ["val"], "field_type": "string", "index_type": "hash"})
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with match")
        query_result = list(query.match("val", "testval1", "testval2").must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1], items[2]]), "Wrong query results")

        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with match and empty result")
        query_result = list(query.match("val", "testval").must_execute())
        # Then ("Check that result is empty")
        assert_that(query_result, empty(), "Wrong query results")

    def test_query_select_dwithin(self, db, namespace, index, rtree_index_and_items):
        # Given("Create namespace with rtree index and items")
        items = rtree_index_and_items
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with dwithin")
        query_result = list(query.dwithin("rtree", Point(3, 2.5), 2.2).must_execute())
        # Then ("Check that selected item is in result")
        expected_items = [items[i] for i in [2, 3, 4]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_equal_position(self, db, namespace, index, array_indexes_and_items):
        # Given("Create namespace with array indexes and items")
        items = array_indexes_and_items
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with equal_position")
        query.where("arr1", CondType.CondEq, 1).where("arr2", CondType.CondEq, 1)
        query_result = list(query.equal_position("arr1", "arr2").must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1]]), "Wrong query results")

    def test_query_select_limit_offset(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with limit and offset")
        query_result = list(query.limit(3).offset(1).must_execute())
        # Then ("Check that selected items are in result")
        expected_items = [items[i] for i in [1, 2, 3]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    @pytest.mark.parametrize("func_total", ["request_total", "cached_total"])
    def test_query_select_request_total(self, db, namespace, index, items, func_total):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with limit and offset")
        query_result = getattr(query, func_total)().limit(1).must_execute()
        # Then ("Check that total is in result")
        assert_that(query_result.total_count(), equal_to(len(items)), "Unexpected total in query results")
        assert_that(query_result.total_count(), not_(equal_to(query_result.count())),
                    "Unexpected total in query results")

    def test_query_select_with_rank(self, db, namespace, index, ft_index_and_items):
        # Given("Create namespace with ft index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query_result = list(query.where("ft", CondType.CondEq, "word~").with_rank().must_execute())
        # Then ("Check that selected item is in result with rank")
        for res_item in query_result:
            assert_that(res_item, has_entries("id", is_(int), "ft", is_(str), "rank()", is_(float)),
                        "Wrong query results")

    def test_query_select_functions(self, db, namespace, index, ft_index_and_items):
        # Given("Create namespace with ft index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query.where("ft", CondType.CondEq, "word~")
        query_result = list(query.functions("ft=highlight(<,>)", "ft=highlight(!!,!)").must_execute())
        # Then ("Check that selected item is in result and highlighted (the first function is applied)")
        query_results_ft = [i["ft"] for i in query_result]
        expected_ft_content = ["one <word>", "<sword> two", "three <work> 333"]
        assert_that(query_results_ft, contains_inanyorder(*expected_ft_content), "Wrong query results")

    def test_query_select_merge(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces with index and items")
        # Given ("Create new query")
        query1 = db.query.new(namespace).where("id", CondType.CondEq, 2)
        # Given ("Create second query")
        query2 = db.query.new(second_namespace).where("id", CondType.CondEq, 1)
        # When ("Make select query with merge")
        query_result = list(query1.merge(query2).must_execute())
        # Then ("Check that selected item is in result with merge applied")
        assert_that(query_result, equal_to([items[2], second_item]), "Wrong query results")

    def test_query_select_explain(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with explain")
        query_result = query.explain().must_execute()
        # Then ("Check that selected items are in result")
        assert_that(list(query_result), equal_to(items), "Wrong query results")
        # Then ("Check that explain is in result")
        explain_results = json.loads(query_result.get_explain_results())
        assert_that(explain_results, has_entries(selectors=is_(list), sort_index="-", total_us=is_(int)),
                    "Wrong explain results")

    @pytest.mark.parametrize("debug_level", LogLevel)
    def test_query_select_debug(self, db, namespace, index, items, debug_level):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query")
        query_result = list(query.debug(debug_level).where("id", CondType.CondEq, 1).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, equal_to([items[1]]), "Wrong query results")

    def test_query_select_strict_mode_empty(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with strict mode")
        query_result = list(query.strict(StrictMode.Empty).where("rand", CondType.CondEq, 1).must_execute())
        # Then ("Check that selected item is in result")
        assert_that(query_result, empty(), "Wrong query results")

    @pytest.mark.parametrize("strict_mode", [StrictMode.NotSet, StrictMode.Names])
    def test_query_select_strict_mode_names_default(self, db, namespace, index, items, strict_mode):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with strict mode")
        query.strict(strict_mode).where("rand", CondType.CondEq, 1)
        err_msg = "Current query strict mode allows filtering by existing fields only. " \
                  f"There are no fields with name 'rand' in namespace '{namespace}'"
        assert_that(calling(query.execute).with_args(), raises(ApiError, pattern=err_msg))

    def test_query_select_strict_mode_indexes(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with strict mode")
        query.strict(StrictMode.Indexes).where("rand", CondType.CondEq, 1)
        err_msg = "Current query strict mode allows filtering by indexes only. " \
                  f"There are no indexes with name 'rand' in namespace '{namespace}'"
        assert_that(calling(query.execute).with_args(), raises(ApiError, pattern=err_msg))

    def test_query_select_wal_any(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with lsn any")
        err_msg = "WAL queries are not supported"
        query.where("#lsn", CondType.CondAny, None)
        assert_that(calling(query.execute).with_args(), raises(ApiError, pattern=err_msg))


class TestQuerySelectAggregations:
    @pytest.mark.parametrize("calculate, aggregate_func", AGGREGATE_FUNCTIONS_MATH)
    def test_query_select_aggregations_math(self, db, namespace, index, items, calculate, aggregate_func):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with aggregations")
        query_result = getattr(query, aggregate_func)("id").must_execute()
        # Then ("Check that result is empty")
        assert_that(list(query_result), empty(), "Wrong query results")
        # Then ("Check aggregation results")
        ids_list = list(range(len(items)))
        expected_agg_result = [{"value": calculate(ids_list), "type": aggregate_func.split("_")[-1], "fields": ["id"]}]
        assert_that(query_result.get_agg_results(), equal_to(expected_agg_result), "Wrong aggregation results")

    @pytest.mark.parametrize("calculate, aggregate_func", AGGREGATE_FUNCTIONS_MATH)
    def test_query_select_aggregations_math_negative(self, db, namespace, index, calculate, aggregate_func):
        # Given("Create namespace with index")
        # Given("Create new items")
        items = [{"id": -i} for i in range(5)]
        for item in items:
            db.item.insert(namespace, item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with aggregations")
        query_result = getattr(query, aggregate_func)("id").must_execute()
        # Then ("Check that result is empty")
        assert_that(list(query_result), empty(), "Wrong query results")
        # Then ("Check aggregation results")
        ids_list = [i["id"] for i in items]
        expected_agg_result = [{"value": calculate(ids_list), "type": aggregate_func.split("_")[-1], "fields": ["id"]}]
        assert_that(query_result.get_agg_results(), equal_to(expected_agg_result), "Wrong aggregation results")

    def test_query_select_distinct(self, db, namespace, index, index_and_duplicate_items):
        # Given("Create namespace with index and duplicate items")
        items = index_and_duplicate_items
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with distinct")
        query_result = query.distinct("idx").must_execute()
        # Then ("Check that only distinct items is in result")
        expected_ids = [0, 1, 3]
        expected_items = [items[i] for i in expected_ids]
        assert_that(list(query_result), equal_to(expected_items), "Wrong query results")
        # Then ("Check aggregation results")
        expected_ids_str = [str(i) for i in expected_ids]
        assert_that(query_result.get_agg_results(),
                    has_item(has_entries(type="distinct", fields=["idx"],
                                         distincts=contains_inanyorder(*expected_ids_str))),
                    "Wrong aggregation results")

    def test_query_select_facet(self, db, namespace, index, index_and_duplicate_items):
        # Given("Create namespace with index and duplicate items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with facet")
        query.aggregate_facet("idx")
        query_result = query.must_execute()
        # Then ("Check that result is empty")
        assert_that(list(query_result), empty(), "Wrong query results")
        # Then ("Check aggregation results")
        expected_facets = [{"count": 1, "values": ["0"]}, {"count": 2, "values": ["1"]}, {"count": 1, "values": ["3"]}]
        assert_that(query_result.get_agg_results(),
                    has_item(has_entries(type="facet", fields=["idx"],
                                         facets=contains_inanyorder(*expected_facets))),
                    "Wrong aggregation results")

    def test_query_select_facet_sort_offset_limit(self, db, namespace, index, index_and_duplicate_items):
        # Given("Create namespace with index and duplicate items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with facet")
        query.aggregate_facet("id", "idx").sort("id", True).limit(2).offset(1)
        query_result = query.must_execute()
        # Then ("Check that result is empty")
        assert_that(list(query_result), empty(), "Wrong query results")
        # Then ("Check aggregation results")
        expected_facets = [{"count": 1, "values": ["2", "1"]}, {"count": 1, "values": ["1", "1"]}]
        assert_that(query_result.get_agg_results(),
                    has_item(has_entries(type="facet", fields=["id", "idx"],
                                         facets=contains_inanyorder(*expected_facets))),
                    "Wrong aggregation results")


class TestQuerySelectSort:
    @pytest.mark.parametrize("is_reversed", [False, True])
    def test_query_select_sort(self, db, namespace, index, items_shuffled, is_reversed):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with sort")
        query_result = list(query.sort("id", is_reversed).must_execute())
        # Then ("Check that selected items are sorted")
        expected_items = sorted(items_shuffled, key=lambda x: x["id"], reverse=is_reversed)
        assert_that(query_result, equal_to(expected_items))

    def test_query_select_sort_expression(self, db, namespace, index, items_shuffled):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with expression sort")
        query_result = list(query.sort("abs(-2 * id) + 1.5").must_execute())
        # Then ("Check that items are sorted")
        expected_items = sorted(items_shuffled, key=lambda x: x["id"])
        assert_that(query_result, equal_to(expected_items))

    @pytest.mark.parametrize("forced_values, expected_ids", [
        (4, [4, 0, 1, 2, 3]),
        ([1, 3], [1, 3, 0, 2, 4])
    ])
    @pytest.mark.parametrize("is_reversed", [False, True])
    def test_query_select_forced_sort(self, db, namespace, index, items_shuffled, forced_values, expected_ids,
                                      is_reversed):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with forced sort")
        query_result = list(query.sort("id", is_reversed, forced_values).must_execute())
        # Then ("Check that selected items are sorted")
        expected_items = [item for i in expected_ids for item in items_shuffled if item["id"] == i]
        if is_reversed:
            expected_items.reverse()
        assert_that(query_result, equal_to(expected_items))

    @pytest.mark.parametrize("is_reversed", [False, True])
    def test_query_select_sort_stpoint(self, db, namespace, index, rtree_index_and_items, is_reversed):
        # Given("Create namespace with rtree index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with sort point distance")
        point = Point(1.1, 1.5)
        query_result = list(query.sort_stpoint_distance("rtree", point, is_reversed).must_execute())
        # Then ("Check that selected items are sorted")
        expected_items = sorted(rtree_index_and_items, key=lambda i: calculate_distance(i["rtree"], [point.x, point.y]),
                                reverse=is_reversed)
        assert_that(query_result, equal_to(expected_items))

    @pytest.mark.parametrize("is_reversed", [False, True])
    def test_query_select_sort_stfield(self, db, namespace, index, is_reversed):
        # Given("Create namespace with index")
        # Given("Create 2 rtree indexes and items")
        db.index.create(namespace, {"name": "rtree1", "json_paths": ["rtree1"], "field_type": "point",
                                    "index_type": "rtree", "rtree_type": "rstar"})
        db.index.create(namespace, {"name": "rtree2", "json_paths": ["rtree2"], "field_type": "point",
                                    "index_type": "rtree", "rtree_type": "rstar"})
        items = [{"id": i, "rtree1": [i, i ** 2], "rtree2": [i, i ** 3 - 4.5]} for i in range(1, 6)]
        for item in items:
            db.item.insert(namespace, item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make select query with sort fields distance")
        query_result = list(query.sort_stfield_distance("rtree1", "rtree2", is_reversed).must_execute())
        # Then ("Check that selected items are sorted")
        expected_items = sorted(items, key=lambda i: calculate_distance(i["rtree1"], i["rtree2"]), reverse=is_reversed)
        assert_that(query_result, equal_to(expected_items))

    def test_query_select_sort_strict_mode_indexes(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Try to select query with strict and sort non-existent index")
        query.strict(StrictMode.Indexes).sort("rand")
        err_msg = "Current query strict mode allows sort by index fields only. " \
                  f"There are no indexes with name 'rand' in namespace '{namespace}'"
        assert_that(calling(query.execute).with_args(), raises(ApiError, pattern=err_msg))


class TestQuerySelectJoin:
    def test_query_select_left_join(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces with index and items")
        # Given ("Create two queries for join")
        query1 = db.query.new(namespace).where("id", CondType.CondLt, 3)
        query2 = db.query.new(second_namespace)
        # When ("Make select query with join")
        query_result = list(query1.join(query2, "joined").on("id", CondType.CondEq, "id").must_execute())
        # Then ("Check that joined item is in result")
        item_with_joined = {"id": 1, "val": "testval1", f"joined_{second_namespace}": [second_item]}
        items[1] = item_with_joined
        assert_that(query_result, equal_to(items[:3]), "Wrong selected items with JOIN")

    def test_query_select_inner_join(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces with index and items")
        # Given ("Create two queries for join")
        query1 = db.query.new(namespace)
        query2 = db.query.new(second_namespace)
        # When ("Make select query with join")
        query_result = list(query1.inner_join(query2, "joined").on("id", CondType.CondEq, "id").must_execute())
        # Then ("Check that joined item is in result")
        item_with_joined = {"id": 1, "val": "testval1", f"joined_{second_namespace}": [second_item]}
        assert_that(query_result, equal_to([item_with_joined]), "Wrong selected items with JOIN")

    def test_query_select_left_and_inner_join(self, db, namespace, index, items, second_namespace, second_items):
        # Given("Create two namespaces with index and items")
        # Given ("Create join query 1")
        query1 = db.query.new(namespace).where("id", CondType.CondRange, [0, 5])
        query2 = db.query.new(second_namespace).where("id", CondType.CondGe, 2)
        join_query1 = query1.left_join(query2, "joined").on("id", CondType.CondEq, "id")
        query3 = db.query.new(second_namespace).where("id", CondType.CondRange, [0, 2])
        join_query2 = join_query1.inner_join(query3, "joined").on("id", CondType.CondEq, "id")
        # When ("Make select query with join")
        query_result = list(join_query2.must_execute())
        # Then ("Check that joined items are in result")
        expected_items = [
            {"id": 1, "val": "testval1", f"joined_2_{second_namespace}": [second_items[0]]},
            {"id": 2, "val": "testval2", f"joined_1_{second_namespace}": [second_items[1]],
             f"joined_2_{second_namespace}": [second_items[1]]}
        ]
        assert_that(query_result, equal_to(expected_items), "Wrong selected items with JOIN")

    def test_cannot_query_select_join_without_on(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces with index and items")
        # Given ("Create two queries for join")
        query1 = db.query.new(namespace)
        query2 = db.query.new(second_namespace)
        # When ("Try to ake select query join without on")
        assert_that(calling(query1.inner_join(query2, "joined").execute).with_args(),
                    raises(ApiError, pattern="Join without ON conditions"))

    def test_query_select_merge_with_joins(self, db, namespace, index, items, second_namespace, second_item):
        # Given("Create two namespaces with index and items")
        # Given ("Create join query 1")
        query11 = db.query.new(namespace)
        query12 = db.query.new(second_namespace)
        join_query1 = query11.inner_join(query12, "joined").on("id", CondType.CondEq, "id")
        # Given ("Create join query 2")
        query21 = db.query.new(namespace).where("id", CondType.CondSet, [0, 1, 3])
        query22 = db.query.new(second_namespace)
        join_query2 = query21.join(query22, "joined").on("id", CondType.CondEq, "id")
        # When ("Make select query with merge")
        query_result = list(join_query1.merge(join_query2).must_execute())
        # Then ("Check that selected items are in result with join and merge applied")
        item_with_joined = {"id": 1, "val": "testval1", f"joined_{second_namespace}": [second_item]}
        expected_items = [item_with_joined, items[0], item_with_joined, items[3]]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_join_with_merges(self, db, namespace, index, items, second_namespace, second_items):
        # Given("Create two namespaces with index and items")
        # Given ("Create merge query 1")
        query11 = db.query.new(namespace).where("id", CondType.CondSet, [2, 3])
        query12 = db.query.new(second_namespace).where("id", CondType.CondEq, 5)
        merge_query1 = query11.merge(query12)
        # Given ("Create merge query 2")
        query21 = db.query.new(second_namespace).where("id", CondType.CondLt, 4).where("id", CondType.CondGe, 2)
        query22 = db.query.new(namespace).where("id", CondType.CondRange, [3, 5])
        merge_query2 = query21.merge(query22)
        # When ("Make select query with join")
        query_result = list(merge_query1.inner_join(merge_query2, "joined").on("id", CondType.CondEq, "id").execute())
        # Then ("Check that selected items are in result with join and merge applied")
        expected_items = [{"id": 2, "val": "testval2",
                           f"joined_{second_namespace}": [{"id": 2, "second_ns_val": "second_ns_testval_2"}]},
                          {"id": 3, "val": "testval3",
                           f"joined_{second_namespace}": [{"id": 3, "second_ns_val": "second_ns_testval_3"}]},
                          {"id": 5, "second_ns_val": "second_ns_testval_5"}]
        assert_that(query_result, equal_to(expected_items), "Wrong query results")

    def test_query_select_sort_and_inner_join(self, db, namespace, index, items, second_namespace, second_items):
        # Given("Create two namespaces with index and items")
        # Given ("Create two queries for join")
        query1 = db.query.new(namespace).sort("id", True)
        query2 = db.query.new(second_namespace)
        # When ("Make select query with sort and join")
        query_result = list(query1.inner_join(query2, "joined").on("id", CondType.CondEq, "id").must_execute())
        # Then ("Check that joined items are in sorted result")
        expected_items = [{"id": i, "val": f"testval{i}", f"joined_{second_namespace}": [second_items[i - 1]]}
                          for i in range(5, 0, -1)]
        assert_that(query_result, equal_to(expected_items), "Wrong selected items with JOIN")

    def test_query_select_inner_join_sort_joined(self, db, namespace, index, items, second_namespace):
        # Given("Create two namespaces with index and items")
        items_2 = [{"id": 1, "age": 1}, {"id": 3, "age": 1}, {"id": 2, "age": 1}]
        [db.item.insert(second_namespace, item) for item in items_2]
        # Given ("Create two queries for join")
        query1 = db.query.new(namespace)
        query2 = db.query.new(second_namespace)
        # When ("Make select query with join and sort")
        query_result = list(query1.inner_join(query2, "joined").on("id", CondType.CondEq, "age")
                            .sort("id").must_execute())
        # Then ("Check that sorted joined items are in result")
        sorted_items2 = sorted(items_2, key=lambda x: x["id"])
        item_with_joined = {"id": 1, "val": "testval1", f"joined_{second_namespace}": sorted_items2}
        assert_that(query_result, equal_to([item_with_joined]), "Wrong selected items with JOIN")


class TestQueryUpdate:
    def test_query_update_set(self, db, namespace, indexes, items):
        # Given("Create namespace with indexes and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make update set query")
        item = random.choice(items)
        modified_item = copy.deepcopy(item)
        modified_item["val"] = "modified"
        query_result = (query.where("id", CondType.CondEq, item["id"])
                        .set("val", ["modified"])
                        .update(timeout=timedelta(milliseconds=1000)))
        # Then ("Check that item is updated")
        assert_that(list(query_result), equal_to([modified_item]), "Wrong update query results after set")
        # Then ("Check that items contain modified and do not contain original item")
        items_after_update = get_ns_items(db, namespace)
        assert_that(items_after_update, has_length(len(items)), "Wrong items count")
        assert_that(items_after_update, has_item(modified_item), "New updated item not is in namespace")
        assert_that(items_after_update, not_(has_item(item)), "Old updated item is in namespace")

    def test_query_update_set_object(self, db, namespace, indexes):
        # Given("Create namespace with index and items")
        # Given ("Create nested index")
        db.index.create(namespace, {"name": "idx", "json_paths": ["nested.field"],
                                    "field_type": "int", "index_type": "hash"})
        # Given ("Create items")
        items = [{"id": i, "nested": {"field": i}} for i in range(3)]
        for item in items:
            db.item.insert(namespace, item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make update set_object query")
        item = random.choice(items)
        modified_item = copy.deepcopy(item)
        modified_item["nested"]["field"] = 10
        query_result = query.where("id", CondType.CondEq, item["id"]).set_object("nested", [{"field": 10}]).update()
        # Then ("Check that item is updated")
        assert_that(list(query_result), equal_to([modified_item]), "Wrong update query results after set object")
        # Then ("Check that items contain modified and do not contain original item")
        items_after_update = get_ns_items(db, namespace)
        assert_that(items_after_update, has_length(len(items)), "Wrong items count")
        assert_that(items_after_update, has_item(modified_item), "New updated item not is in namespace")
        assert_that(items_after_update, not_(has_item(item)), "Old updated item is in namespace")

    def test_query_update_expression(self, db, namespace, index, items):
        # Given("Create namespace with indexes and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make update expression query")
        item = random.choice(items)
        modified_item = copy.deepcopy(item)
        modified_item["field"] = item["id"] + 0.5
        query_result = query.where("id", CondType.CondEq, item["id"]).expression("field", "id + 0.5*serial()").update()
        # Then ("Check that item is updated")
        assert_that(list(query_result), equal_to([modified_item]), "Wrong update query results after set")
        # Then ("Check that items contain modified and do not contain original item")
        items_after_update = get_ns_items(db, namespace)
        assert_that(items_after_update, has_length(len(items)), "Wrong items count")
        assert_that(items_after_update, has_item(modified_item), "New updated item not is in namespace")
        assert_that(items_after_update, not_(has_item(item)), "Old updated item is in namespace")

    def test_query_drop(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make update drop query")
        item = random.choice(items)
        query_result = list(query.where("id", CondType.CondEq, item["id"]).drop("val").update())
        # Then ("Check that result contains item with dropped field")
        del item["val"]
        assert_that(query_result, equal_to([item]), "Wrong update query results after drop")
        # Then ("Check that field was dropped for a chosen item")
        items_after_drop = get_ns_items(db, namespace)
        assert_that(items_after_drop, equal_to(items), "Wrong items after drop")

    def test_query_drop_all(self, db, namespace, index, sparse_index, items):
        # Given("Create namespace with index")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make update drop query")
        query_result = list(query.drop("val").update())
        # Then ("Check that result contains items with dropped field")
        for item in items:
            del item["val"]
        assert_that(query_result, equal_to(items), "Wrong update query results after drop")
        # Then ("Check that field was dropped")
        items_after_drop = get_ns_items(db, namespace)
        assert_that(items_after_drop, equal_to(items), "Wrong items after drop")

    def test_cannot_query_drop_not_sparse(self, db, namespace, indexes, items):
        # Given("Create namespace with indexs and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Try to make update drop query with not sparse index")
        assert_that(calling(query.drop("val").update).with_args(),
                    raises(Exception,
                           pattern="It's only possible to drop sparse or non-index fields via UPDATE statement!"))


class TestQueryDelete:
    def test_query_delete(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make delete query")
        item = random.choice(items)
        query_result = query.where("id", CondType.CondEq, item["id"]).delete(timeout=timedelta(milliseconds=1000))
        # Then ("Check that chosen item was deleted")
        assert_that(query_result, equal_to(1), "Wrong delete items count")
        items_after_delete = get_ns_items(db, namespace)
        assert_that(items_after_delete, has_length(len(items) - 1), "Wrong items count after delete")
        assert_that(items_after_delete, not_(has_item(item)), "Deleted item is in namespace")

    def test_query_delete_all(self, db, namespace, index, items):
        # Given("Create namespace with index and items")
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Make delete query")
        item = random.choice(items)
        query_result = query.delete()
        # Then ("Check that chosen item was deleted")
        assert_that(query_result, equal_to(10), "Wrong delete items count")
        items_after_delete = get_ns_items(db, namespace)
        assert_that(items_after_delete, empty(), "Wrong items count after delete")


class TestQueryTimeouts:
    def test_query_select_timeout_small(self, db, namespace, index):
        # Given("Create namespace with items")
        items = [{"id": i, "val": f"testval{i}", "non_idx": i + 1} for i in range(10000)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        q1 = (db.query.new(namespace).where("id", CondType.CondGt, -1)
              .where("non_idx", CondType.CondGt, 0))
        q2 = (db.query.new(namespace).where("val", CondType.CondLt, "testval1000")
              .where("non_idx", CondType.CondGt, 0))
        q3 = (db.query.new(namespace).where("val", CondType.CondGt, "testval1000")
              .where("id", CondType.CondRange, [1, 9000])
              .where("non_idx", CondType.CondGt, 100))
        q4 = (db.query.new(namespace).where("non_idx", CondType.CondGt, "id")
              .where("id", CondType.CondEq, [1, 9000]).op_not().where("id", CondType.CondLt, 100))
        query = q1.merge(q2).merge(q3).merge(q4)
        # When ("Try to make select query with small timeout")
        assert_that(calling(query.execute).with_args(timeout=timedelta(milliseconds=1)),
                    raises(ApiError, pattern="Context timeout|Read lock (.*) was canceled or timed out (mutex)"))

    def test_query_select_join_timeout_small(self, db, namespace, index):
        # Given("Create namespace with items")
        items = [{"id": i, "val": f"testval{i}"} for i in range(2500)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query with join")
        query1 = db.query.new(namespace).where("id", CondType.CondGt, 0)
        query2 = db.query.new(namespace).where("id", CondType.CondGe, 1)
        query = query1.join(query2, "joined").on("val", CondType.CondGt, "val")
        # When ("Try to make select query with small timeout")
        assert_that(calling(query.execute).with_args(timeout=timedelta(milliseconds=1)),
                    raises(ApiError, pattern="Context timeout|Read lock (.*) was canceled or timed out (mutex)"))

    @pytest.mark.skip(reason="1951")
    def test_query_update_timeout_small(self, db, namespace, index):
        # Given("Create namespace with items")
        items = [{"id": i, "val": f"testval{i}"} for i in range(15000)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace).set("val", ["modified"])
        # When ("Try to make update set query with small timeout")
        assert_that(calling(query.update).with_args(timeout=timedelta(milliseconds=1)),
                    raises(ApiError, pattern="Context timeout|Read lock (.*) was canceled or timed out (mutex)"))
        # Then ("Check that items were not updated")
        items_after_update = get_ns_items(db, namespace)
        assert_that(items_after_update, equal_to(items), "Items were updated")


class TestQueryKNN:

    def test_query_knn_param_negative(self, db, namespace):
        # Given ("Create test random float vector")
        vec = random_vector(2)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Check indexes search param")
        assert_that(calling(IndexSearchParamBruteForce),
                    raises(ValueError, pattern="Either 'k' or 'radius' needs to be specified"))
        assert_that(calling(IndexSearchParamBruteForce).with_args(k=0, radius=0.0),
                    raises(ValueError, pattern="KNN limit 'k' should not be less than 1"))
        assert_that(calling(IndexSearchParamHnsw).with_args(ef=0),
                    raises(ValueError, pattern="Either 'k' or 'radius' needs to be specified"))
        assert_that(calling(IndexSearchParamHnsw).with_args(k=0, ef=1),
                    raises(ValueError, pattern="KNN limit 'k' should not be less than 1"))
        assert_that(calling(IndexSearchParamHnsw).with_args(k=2, ef=1),
                    raises(ValueError, pattern="'ef' should not be less than 'k'"))
        assert_that(calling(IndexSearchParamIvf).with_args(nprobe=0),
                    raises(ValueError, pattern="Either 'k' or 'radius' needs to be specified"))
        assert_that(calling(IndexSearchParamIvf).with_args(k=0, nprobe=1),
                    raises(ValueError, pattern="KNN limit 'k' should not be less than 1"))
        assert_that(calling(IndexSearchParamIvf).with_args(k=1, nprobe=0),
                    raises(ValueError, pattern="'nprobe' should not be less than 1"))
        assert_that(calling(IndexSearchParamIvf).with_args(radius=1, nprobe=0),
                    raises(ValueError, pattern="'nprobe' should not be less than 1"))
        assert_that(calling(IndexSearchParamIvf).with_args(k=1, radius=2, nprobe=0),
                    raises(ValueError, pattern="'nprobe' should not be less than 1"))

        # When ("Make query with knn")
        assert_that(calling(query.where_knn).with_args("vec", None, None),
                    raises(QueryError, pattern="A required parameter is not specified. `vec` can't be None or empty"))
        assert_that(calling(query.where_knn).with_args("vec", [], None),
                    raises(QueryError, pattern="A required parameter is not specified. `vec` can't be None or empty"))
        assert_that(calling(query.where_knn).with_args("vec", vec, None),
                    raises(QueryError, pattern="A required parameter is not specified. `param` can't be None"))

    def test_query_brute_force(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_bf)
        index["config"] = {"dimension": dimension, "metric": "inner_product", "start_size": 1000}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        k: Final[int] = 39
        param = IndexSearchParamBruteForce(k=k)
        query_result = list(
            query.where_knn("vec", random_vector(dimension), param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)
        assert_that(query_result, has_length(k))

    def test_query_hnsw(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_hnsw)
        index["config"] = {"dimension": dimension, "metric": "inner_product", "start_size": 1000, "m": 16,
                           "ef_construction": 200}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamHnsw(k=30, ef=30)
        query_result = list(
            query.where_knn("vec", random_vector(dimension), param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    def test_query_hnsw_multithread(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 8
        index = copy.deepcopy(vector_index_hnsw)
        index["config"] = {"dimension": dimension, "metric": "inner_product", "start_size": 1000, "m": 16,
                           "ef_construction": 200, "multithreading": 1}
        db.index.create(namespace, index)
        # Given("Insert items via transaction")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        transaction = db.tx.begin(namespace)
        for item in items:
            transaction.insert_item(item)
        transaction.commit()
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamHnsw(k=25, ef=40)
        query_result = list(
            query.where_knn("vec", random_vector(dimension), param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    def test_query_ivf(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 2
        index = copy.deepcopy(vector_index_ivf)
        index["config"] = {"dimension": dimension, "metric": "l2", "centroids_count": 3}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": random_vector(dimension)} for i in range(100)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamIvf(k=30, nprobe=2)
        query_result = list(
            query.where_knn("vec", random_vector(dimension), param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_close_to_ns_items(query_result, items)

    def test_query_bf_with_radius_in_index_config(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 2
        index = copy.deepcopy(vector_index_bf)
        index["config"] = {"dimension": dimension, "metric": "l2", "radius": 20.0}
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": [1, i]} for i in range(20)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamBruteForce(k=30)
        query_result = list(
            query.where_knn("vec", [3, 4], param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_only_close_to_items(query_result, items[1:8])

    def test_query_hnsw_with_radius_in_config_and_query(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 2
        index = copy.deepcopy(vector_index_hnsw)
        index["config"]["dimension"] = dimension
        index["config"]["metric"] = "inner_product"
        index["config"]["radius"] = 21.0  # will not be applied
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": [1, i]} for i in range(20)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query")
        param = IndexSearchParamHnsw(ef=30, radius=50)
        query_result = list(
            query.where_knn("vec", [5, 5], param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_only_close_to_items(query_result, items[10:])

    def test_query_ivf_with_k_and_radius_in_query(self, db, namespace, index):
        # Given ("Create float vector index")
        dimension: Final[int] = 2
        index = copy.deepcopy(vector_index_ivf)
        index["config"]["dimension"] = dimension
        index["config"]["metric"] = "cosine"
        db.index.create(namespace, index)
        # Given("Insert items")
        items = [{"id": i, "vec": [1, i]} for i in range(20)]
        for item in items:
            db.item.insert("new_ns", item)
        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query with limit by radius")
        param = IndexSearchParamIvf(k=100, nprobe=2, radius=0.87)
        query_result = list(
            query.where_knn("vec", [3, 4], param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_only_close_to_items(query_result, items[1:8])

        # Given ("Create new query")
        query = db.query.new(namespace)
        # When ("Execute query with limit by K")
        param = IndexSearchParamIvf(k=3, nprobe=2, radius=0.87)
        query_result = list(
            query.where_knn("vec", [3, 4], param)
            .select("vectors()")
            .execute(timeout=timedelta(seconds=1)))
        # Then ("Check knn select result")
        check_response_has_only_close_to_items(query_result, items[1:4])
